/******
        file     : main_shared_motifs.c
        function : align 2 sequence by an algo derived from Waterman and Eggert, J Mol. Biol. (1987)
                   the algo is described in Castillo-Davis, Hartl and Achaz (manuscript submitted)
                   please send all comments to gachaz@oeb.harvard.edu
                                                 
        created  : 8 Mar 2003
        modif    : 6 May 03, 13 Jan04, Aug 04 (change readfst), Sep 2004 (improve help)

        author   : amikezor <G. Achaz>
*****/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>

#include "SharMot.h"

#define _VERSION_ "3.3 (Sep 2004)"
#define ALPHABET "ACGTNX"


static void printusage(char *arg0)
{
	fprintf(stderr,"Syntax is '%s [-hv] [-AMTDS] [-o #] [-e #] -l # file1.fst file2.fst'\n", arg0);
}

static void printversion(char *arg0)
{
	printf("** Shared Motifs - version %s **\n",_VERSION_);
}


static void printhelp(char *arg0)
{
	printversion(arg0);
 	fprintf(stderr,"** detect all Shared Motifs between 2 sequences **\n");
 	fprintf(stderr,"** recursive local alignment in both directions **\n\n");
	
 	printusage(arg0);
  fprintf(stderr,"\n"
  "  Necessary:\n"
  "\t-l # : set the minimum 'l'ength of perfect strech of matches\n"
  "\t       This parameter is crucial and should be chosen carefully.\n"
  "\t       The actual minimum score will be 4*l (since a match is +4)\n"
  "\t       Note that for scores, 15bp with 0 mismatch = 17bp with 1 mismatch\n"
  "  Optionnal:\n"
  "\t-S   : print out 'S'hared Motifs\n"
  "\t-A   : print out 'A'lignments \n"
  "\t-M   : print out 'M'atrices (alignment matrices --scores and traces--)\n"
  "\t-T   : print out 'T'ableR (number of times each position is in the other sequence) \n"
  "\t-D   : print out 'D'otplot (one dot for each match).\n"
  "\t-o # : set the gap_'o'pen value, as the number of times a match score (default is 4)\n"
  "\t-e # : set the gap_'e'xt value, as the number of times a match score (default is 4)\n"
  "  Information:\n"
  "\t-h   : Guess...\n"
  "\t-v   : print 'v'ersion and exit\n"
  );

   fprintf(stderr,"Files:\n");
   fprintf(stderr,"\tSequences should be in a standard fasta format\n");
	fprintf(stderr,"\tAny non-A/C/G/T/X symbol is treated as an N\n");
   fprintf(stderr,"\tAll Xs are considered as un-alignable and can be used to mask\n\n");

	fprintf(stderr,"Alignments Scores:\n");
	fprintf(stderr,"\tMatch(A|C|G|T)=+4; Match(Ni)=+1; Match(Xi)=-100,000; Mismatch=-4;\n\n");

  	fprintf(stderr,"The algorithm is derived from Waterman & Eggert, J Mol. Biol. (1987). It is\n"); 
 	fprintf(stderr,"described in Castillo-Davis, Hartl & Achaz, Genome Res. (2004). Please\n");
 	fprintf(stderr,"cite the previous reference if you use this software for publication.\n");
 	fprintf(stderr,"Please send any correspondance (comment/bug/help) to <gachaz@oeb.harvard.edu>\n");

}


int main(int argc, char **argv){

	char *file1, *file2;           /* sequence files in fst format */
	
	char *seq1, *seq2,             /* sequences in strings format */
	     *seq2inv, *BigSeq=NULL;   /* sequence for inverted and seq1+seq2 for matrixlog */

	long size1, size2;             /* size of seq1 and seq2 */
		
	int gap_open=4,                /* gap open penalty will be : GAP_OPEN = - gap_open*mean_score */
	     gap_ext=4;                /* gap extension penalty will be : GAP_EXT = - gap_ext*mean_score  */

	SCORING Scoring;               /* to store all scoring for alignment */
	RESULTS Results;               /* to store resukts for each alignemnt */

	int *tableR1, *tableR2;        /* table of Redundancy of seq1 and seq2 */

	int  minimum_strech=0,         /* defined by the user as the length of perfect strech to be considered */
        minimum_score;            /* rescale the minimum_strech into our scroing scale */
		  
	int round=0;                   /* the number of suboptimal alignment */
	
	short opt_Matrix=0,            /* print out Matrices */
	      opt_align=0,             /* Alignemnt */
			opt_tableR=0,            /* tableR */ 
			opt_dotplot=0,           /* print out dotplot */ 
			opt_SharMot=0;           /* print out shared motifs */ 

	extern char *optarg;           /* for options parisng */
	extern  int optind;
	int c;

	int i,
		count1=0,                   /* dummy variables */
		count2=0;

	/*
		Parse the options
	*/

	while((c = getopt(argc, argv, "hvMATSDo:e:l:")) != -1)
		switch(c)
		{
			case 'h' :
				printhelp(argv[0]);
				exit(1);

			case 'v' :
				printversion(argv[0]);
				exit(1);

			case 'M' :
				opt_Matrix=1;
				break;

			case 'S' :
				opt_SharMot=1;
				break;

			case 'A' :
				opt_align=1;
				break;

			case 'D' :
				opt_dotplot=1;
				break;

			case 'T' :
				opt_tableR=1;
				break;

			case 'o' :
				gap_open = atoi(optarg);
				break;

			case 'e' :
				gap_ext  = atoi(optarg);
				break;

			case 'l':
				minimum_strech=atoi(optarg);
				break;

			case '?':
				printusage(argv[0]),exit(1);
	}


	/*
		Check and get the remaining arguments
	*/	
	if(argc-optind != 2 || !minimum_strech )
		printusage(argv[0]),exit(1);

 	file1 = argv[optind];
	file2 = argv[optind+1];

	/*
		Get sequences and check and clean them
	*/
	seq1 = file2mem(file1, & size1);
	seq2 = file2mem(file2, & size2);

	UpperSequence(seq1);
	UpperSequence(seq2);
	
	CheckSeq(seq1, ALPHABET);
	CheckSeq(seq2, ALPHABET);
	
	nonACGTXtoN(seq1);
	nonACGTXtoN(seq2);

	seq2inv = malloc( (size_t) (size2+1)*sizeof(char));
	if(!seq2inv)fprintf(stderr,"main: malloc_error\n"),exit(3);
	invseq(seq2, seq2inv);
	
	/*
		Get matrices - identity 
	*/
	
	BigSeq = (char *)malloc( (size1+size2+1) * sizeof(char) ) ;
	if(!BigSeq)fprintf(stderr,"main: malloc_error\n"),exit(3);
	strcpy(BigSeq,seq1);
	strcpy(BigSeq+size1,seq2);
	
	identity_matrix(&Scoring, gap_open, gap_ext);
		
	mem_RecursivAlign( &Results, seq1, seq2);
	mem_tableR(size1, size2, &tableR1, &tableR2);
	
	/*
		set it to the mean_match*minimumscore (e.g. a strech of perfect match that size)
	*/
	minimum_score = minimum_strech*4*Scoring.matrix[CHAR2SYMB('N')][CHAR2SYMB('N')];
	
	/*
		print out headers
	*/
	printf("\n# Shared motifs detection version %s\n",_VERSION_);
	printf("# LenSeq1: %ld ; LenSeq2: %ld ; Matrix : %s ; MinumScore: %d\n",size1, size2, "Identity",minimum_score );

	if(opt_SharMot)
		printf("# orientation\tpos1\tpos2\tlen1\tlen2\tmatches\tscore\n");
	
	/*
		Do one orientation
	*/
	do{
	
		waterman_eggert(seq1, seq2, &Scoring, &Results, round, opt_Matrix);
		
		if(Results.BestScore <  minimum_score)	break;
				
		Update_tableR(tableR1, tableR2, Results.Begin_row-1, Results.BestScore_row-Results.Begin_row+1,
		                                Results.Begin_col-1, Results.BestScore_col-Results.Begin_col+1); 
		
		/*
			Print out results
		*/
		if(opt_align)print_alignment( &Results , &Scoring, seq1, seq2);
		if(opt_dotplot)print_dotplot( &Results , "direct");
		if(opt_SharMot)print_shared_motif( &Results , &Scoring, "direct");
		
		round++;
		
	}while(1);
	
	/*
		Reset some values
	*/
	round=0;
	
	strcpy(BigSeq+size1,seq2inv);
	
	minimum_score = minimum_strech*4*Scoring.matrix[CHAR2SYMB('N')][CHAR2SYMB('N')]; 
		
	memset_RecursivAlign(&Results, seq1, seq2);
			
	/*
		do the other one
	*/
	do{
	
		waterman_eggert(seq1, seq2inv, &Scoring, &Results, round, opt_Matrix);
		
		if(Results.BestScore <  minimum_score)
			break;
			
		Update_tableR(tableR1, tableR2, Results.Begin_row-1, Results.BestScore_row-Results.Begin_row+1,
		                                Results.size_col-Results.BestScore_col-1, Results.BestScore_col-Results.Begin_col+1); 

		/*
			Print out results
		*/
		if(opt_align)print_alignment( &Results , &Scoring, seq1, seq2inv);
		if(opt_dotplot)print_dotplot( &Results , "inverted");
		if(opt_SharMot)print_shared_motif( &Results , &Scoring, "inverted");

		round++;
		
	}while(1);
	
	Finish_tableR(size1, size2, tableR1, tableR2);
	
	
	/*
		print out tableRs
	*/
	printf("\n# Proportions of shared motifs\n");
	if( opt_tableR && printf("# TableR Seq1\n") )
		for(i=0;i<size1;i++){
			if(tableR1[i]){
				printf("tableR1 %d %d\n",i+1,tableR1[i]);
				count1++;
			}
		}
	else{
		for(count1=0,i=0;i<size1;i++)if(tableR1[i])count1++;
		printf("seq1 shares %d on %ld bp (%.3f %%)\n",count1,size1,100*(float)count1/(float)size1);
	}
	
	if( opt_tableR && printf("# TableR Seq2\n"))
		for(i=0;i<size2;i++){
			if(tableR2[i]){
				printf("tableR2 %d %d\n",i+1,tableR2[i]);
				count2++;
			}
		}
	else{
		for(count2=0,i=0;i<size2;i++)if(tableR2[i])count2++;
		printf("seq2 shares %d on %ld bp (%.3f %%)\n",count2,size2,100*(float)count2/(float)size2);
	}

	printf("SMd = %f\n", (float)(size1+size2 - count1-count2)/(float)(size1+size2) );

	/*
		Free everyone
	*/

	free_RecursivAlign(&Results);
	free_tableR(tableR1,tableR2);

	if(BigSeq)free(BigSeq);
	free(seq1);
	free(seq2);
	free(seq2inv);
	
	return 0;

}

